import pytest
import subprocess
import importlib.metadata
from unittest.mock import patch
from pydepguard.pylock.validator import resolve_installed_package_info, check_package_availability, validate_environment

def test_importlib_succeeds(monkeypatch):
    class FakeMetadata:
        @staticmethod
        def version(name):
            if name == "requests":
                return "2.31.0"
            raise importlib.metadata.PackageNotFoundError

    monkeypatch.setattr("importlib.metadata.version", FakeMetadata.version)

    result = resolve_installed_package_info("requests")
    assert result['available'] is True
    assert result['version'] == "2.31.0"
    assert result['source'] == "importlib"

def test_importlib_fails_pip_show_succeeds(monkeypatch):
    def fail_version(name):
        raise importlib.metadata.PackageNotFoundError

    monkeypatch.setattr("importlib.metadata.version", fail_version)

    class FakeCompletedProcess:
        def __init__(self):
            self.returncode = 0
            self.stdout = b"Name: example\nVersion: 1.2.3\n"
            self.stderr = b""

    monkeypatch.setattr(subprocess, "run", lambda *a, **kw: FakeCompletedProcess())

    result = resolve_installed_package_info("example")
    assert result['available'] is True
    assert result['version'] == "1.2.3"
    assert result['source'] == "pip_show"

def test_pip_show_no_version(monkeypatch):
    def fail_version(name):
        raise importlib.metadata.PackageNotFoundError

    monkeypatch.setattr("importlib.metadata.version", fail_version)

    class NoVersionProcess:
        def __init__(self):
            self.returncode = 0
            self.stdout = b"Name: something\n"
            self.stderr = b""

    monkeypatch.setattr(subprocess, "run", lambda *a, **kw: NoVersionProcess())

    result = resolve_installed_package_info("something")
    assert result['available'] is True
    assert result['version'] is None
    assert result['source'] == "pip_show"

def test_all_methods_fail(monkeypatch):
    def fail_version(name):
        raise importlib.metadata.PackageNotFoundError

    class FakeProcess:
        def __init__(self):
            self.returncode = 1
            self.stdout = b""
            self.stderr = b""

    monkeypatch.setattr("importlib.metadata.version", fail_version)
    monkeypatch.setattr(subprocess, "run", lambda *a, **kw: FakeProcess())

    result = resolve_installed_package_info("nonexistent")
    assert result['available'] is False
    assert result['version'] is None
    assert result['source'] == "none"

def test_version_match(monkeypatch):
    monkeypatch.setattr("importlib.metadata.version", lambda _: "2.0.0")
    result = check_package_availability("flask", expected_version="2.0.0")
    assert result['available'] is True
    assert result['version_matches'] is True

def test_version_mismatch(monkeypatch):
    monkeypatch.setattr("importlib.metadata.version", lambda _: "2.1.0")
    result = check_package_availability("flask", expected_version="2.0.0")
    assert result['available'] is True
    assert result['version_matches'] is False

def test_validate_passes_with_defaults(monkeypatch):
    monkeypatch.setattr("pydepguard.pylock.validator.check_package_availability", lambda dep, ver=None: {
        'available': True,
        'version_matches': True,
        'version': ver,
        'source': 'importlib'
    })

    lockfile = {'deps': {'flask': {'version': '2.0.0'}}}
    validate_environment(lockfile)


def test_validate_missing_package_warn(monkeypatch):
    monkeypatch.setattr("pydepguard.pylock.validator.check_package_availability", lambda dep, ver=None: {
        'available': False,
        'version_matches': False,
        'version': None,
        'source': 'none'
    })

    lockfile = {'deps': {'flask': {'version': '2.0.0'}}}
    validate_environment(lockfile, on_error='warn', interactive=False)


def test_validate_version_mismatch_strict(monkeypatch):
    monkeypatch.setattr("pydepguard.pylock.validator.check_package_availability", lambda dep, ver=None: {
        'available': True,
        'version_matches': False,
        'version': '1.0.0',
        'source': 'importlib'
    })

    lockfile = {'deps': {'flask': {'version': '2.0.0'}}}
    with pytest.raises(RuntimeError, match="Validation failed due to version mismatch"):
        validate_environment(lockfile, interactive=False)


def test_validate_version_mismatch_non_strict(monkeypatch):
    monkeypatch.setattr("pydepguard.pylock.validator.check_package_availability", lambda dep, ver=None: {
        'available': True,
        'version_matches': False,
        'version': '1.0.0',
        'source': 'importlib'
    })

    lockfile = {'deps': {'flask': {'version': '2.0.0'}}}
    validate_environment(lockfile, strict=False, interactive=False)


def test_validate_malformed_lockfile():
    with pytest.raises(ValueError, match="Invalid lockfile format"):
        validate_environment({})


def test_validate_check_package_crashes_abort(monkeypatch):
    def raise_error(*args, **kwargs):
        raise Exception("Simulated failure")
    monkeypatch.setattr("pydepguard.pylock.validator.check_package_availability", raise_error)
    lockfile = {'deps': {'broken': {'version': '1.0.0'}}}
    with pytest.raises(RuntimeError, match="Dependency check failed for broken"):
        validate_environment(lockfile, on_error='abort')


def test_validate_check_package_crashes_warn(monkeypatch):
    def raise_error(*args, **kwargs):
        raise Exception("Simulated failure")
    monkeypatch.setattr("pydepguard.pylock.validator.check_package_availability", raise_error)
    lockfile = {'deps': {'broken': {'version': '1.0.0'}}}
    validate_environment(lockfile, on_error='warn')  # Should not raise


def test_validate_missing_package_skip(monkeypatch):
    monkeypatch.setattr("pydepguard.pylock.validator.check_package_availability", lambda dep, ver=None: {
        'available': False,
        'version_matches': False,
        'version': None,
        'source': 'none'
    })
    lockfile = {'deps': {'flask': {'version': '2.0.0'}}}
    validate_environment(lockfile, on_error='skip')  # Should not raise


def test_validate_version_mismatch_input_decline(monkeypatch):
    monkeypatch.setattr("pydepguard.pylock.validator.check_package_availability", lambda dep, ver=None: {
        'available': True,
        'version_matches': False,
        'version': '1.0.0',
        'source': 'importlib'
    })
    monkeypatch.setattr("builtins.input", lambda _: "no")
    lockfile = {'deps': {'flask': {'version': '2.0.0'}}}
    with pytest.raises(RuntimeError, match="Validation aborted due to version mismatch."):
        validate_environment(lockfile, strict=True, interactive=True)


def test_validate_version_mismatch_keyboard_interrupt(monkeypatch):
    monkeypatch.setattr("pydepguard.pylock.validator.check_package_availability", lambda dep, ver=None: {
        'available': True,
        'version_matches': False,
        'version': '1.0.0',
        'source': 'importlib'
    })
    monkeypatch.setattr("builtins.input", lambda _: (_ for _ in ()).throw(KeyboardInterrupt))
    lockfile = {'deps': {'flask': {'version': '2.0.0'}}}
    with pytest.raises(SystemExit) as e:
        validate_environment(lockfile, strict=True, interactive=True)
    assert e.value.code == 130

def test_validate_check_package_crashes_skip(monkeypatch):
    def raise_error(*args, **kwargs):
        raise Exception("Simulated failure")
    monkeypatch.setattr("pydepguard.pylock.validator.check_package_availability", raise_error)

    lockfile = {'deps': {'broken': {'version': '1.0.0'}}}
    # Should reach the final else branch under crash condition
    validate_environment(lockfile, on_error='skip')


def test_validate_missing_package_abort(monkeypatch):
    monkeypatch.setattr("pydepguard.pylock.validator.check_package_availability", lambda dep, ver=None: {
        'available': False,
        'version_matches': True,
        'version': None,
        'source': 'mock'
    })
    lockfile = {'deps': {'missing': {'version': '1.0.0'}}}
    with pytest.raises(RuntimeError, match="Missing required package: missing"):
        validate_environment(lockfile, on_error='abort')


def test_validate_missing_package_else_skip(monkeypatch):
    monkeypatch.setattr("pydepguard.pylock.validator.check_package_availability", lambda dep, ver=None: {
        'available': False,
        'version_matches': True,
        'version': None,
        'source': 'mock'
    })
    lockfile = {'deps': {'missing': {'version': '1.0.0'}}}
    validate_environment(lockfile, on_error='skip')


def test_validate_version_mismatch_warn(monkeypatch):
    monkeypatch.setattr("pydepguard.pylock.validator.check_package_availability", lambda dep, ver=None: {
        'available': True,
        'version_matches': False,
        'version': '0.9.0',
        'source': 'mock'
    })
    lockfile = {'deps': {'mismatch': {'version': '1.0.0'}}}
    validate_environment(lockfile, strict=True, interactive=False, on_error='warn')


def test_validate_version_mismatch_else_skip(monkeypatch):
    monkeypatch.setattr("pydepguard.pylock.validator.check_package_availability", lambda dep, ver=None: {
        'available': True,
        'version_matches': False,
        'version': '0.9.0',
        'source': 'mock'
    })
    lockfile = {'deps': {'mismatch': {'version': '1.0.0'}}}
    validate_environment(lockfile, strict=True, interactive=False, on_error='skip')




import ast
import builtins
from pathlib import Path
from dataclasses import dataclass

@dataclass
class ImportReference:
    module: str
    file: str
    line: int
    import_type: str
    imported_symbols: list[str] = None

@dataclass
class SymbolReference:
    name: str
    file: str
    line: int
    context: str

def scan_script_for_imports(filepath: Path) -> tuple[list[ImportReference], list[SymbolReference]]:
    refs = []
    used_references = []
    declared_symbols = set()
    BUILTIN_SYMBOLS = set(dir(builtins))

    with open(filepath, 'r', encoding='utf-8') as f:
        try:
            tree = ast.parse(f.read(), filename=str(filepath))
        except SyntaxError as e:
            print(f"[ERROR] Failed to parse {filepath}: {e}")
            return refs, []

    for node in ast.walk(tree):
        if isinstance(node, ast.Import):
            for alias in node.names:
                refs.append(ImportReference(
                    module=alias.name,
                    file=str(filepath),
                    line=node.lineno,
                    import_type='import',
                    imported_symbols=[alias.asname] if alias.asname else []
                ))
                declared_symbols.add(alias.asname or alias.name.split('.')[0])

        elif isinstance(node, ast.ImportFrom):
            if node.level and not node.module:
                module = '.' * node.level
                symbols = [alias.name for alias in node.names]
            else:
                module = f"{'.' * node.level}{node.module}" if node.level else node.module
                symbols = [alias.name for alias in node.names]

            refs.append(ImportReference(
                module=module,
                file=str(filepath),
                line=node.lineno,
                import_type='from',
                imported_symbols=symbols
            ))
            declared_symbols.update(symbols)

        elif isinstance(node, ast.Call):
            if isinstance(node.func, ast.Name) and node.func.id == '__import__':
                if len(node.args) >= 1 and isinstance(node.args[0], ast.Constant) and isinstance(node.args[0].value, str):
                    refs.append(ImportReference(
                        module=node.args[0].value,
                        file=str(filepath),
                        line=node.lineno,
                        import_type='dynamic',
                        imported_symbols=[]
                    ))
                    declared_symbols.add(node.args[0].value)

            elif isinstance(node.func, ast.Attribute) and node.func.attr == 'import_module':
                if hasattr(node.func.value, 'id') and node.func.value.id == 'importlib':
                    if len(node.args) >= 1 and isinstance(node.args[0], ast.Constant) and isinstance(node.args[0].value, str):
                        refs.append(ImportReference(
                            module=node.args[0].value,
                            file=str(filepath),
                            line=node.lineno,
                            import_type='dynamic',
                            imported_symbols=[]
                        ))
                        declared_symbols.add(node.args[0].value)

        elif isinstance(node, ast.Name):
            if isinstance(node.ctx, ast.Load):
                used_references.append(SymbolReference(
                    name=node.id,
                    file=str(filepath),
                    line=node.lineno,
                    context='load'
                ))

        elif isinstance(node, ast.Attribute):
            if isinstance(node.value, ast.Name):
                used_references.append(SymbolReference(
                    name=node.value.id,
                    file=str(filepath),
                    line=node.lineno,
                    context='attribute'
                ))

        elif isinstance(node, ast.Assign):
            for target in node.targets:
                if isinstance(target, ast.Name):
                    declared_symbols.add(target.id)

        elif isinstance(node, ast.With):
            for item in node.items:
                if item.optional_vars and isinstance(item.optional_vars, ast.Name):
                    declared_symbols.add(item.optional_vars.id)

    imported_modules = {ref.module.split('.')[0] for ref in refs}
    imported_aliases = {
        sym for ref in refs if ref.imported_symbols
        for sym in ref.imported_symbols
    }
    all_imported = imported_modules.union(imported_aliases)
    all_known = all_imported.union(declared_symbols).union(BUILTIN_SYMBOLS)

    seen = set()
    unbound_symbols = []
    for ref in used_references:
        if ref.name not in all_known and ref.name not in seen:
            unbound_symbols.append(ref)
            seen.add(ref.name)

    return refs, unbound_symbols



import subprocess
import sys
import importlib.metadata
import importlib.resources as resources
import json
import re
from .cache import KNOWN_DEP_MAP



def guess_distribution_name(module_name: str):
    for dist in importlib.metadata.distributions():
        try:
            top_levels = dist.read_text("top_level.txt")
            if top_levels:
                names = [line.strip() for line in top_levels.splitlines()]
                if module_name in names:
                    return dist.metadata["Name"]
        except Exception:
            continue
    return None


def install_package(package: str, version: str = None, _is_retry=False):

    mapped = KNOWN_DEP_MAP.get(package.lower())

    if mapped and mapped.lower() != package.lower():
        print(f"[pylock] Using mapped pip name: {package} → {mapped}")
        package = mapped
        version = ""

    if version == "unknown":
        version = ""

    pkg = f"{package}=={version}" if version else package

    print(f"[pylock] Installing {pkg} ...")

    result = subprocess.run(
        [sys.executable, "-m", "pip", "install", pkg],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )

    if result.returncode == 0:
        version = extract_installed_version(result.stdout, package)

        print(f"[pylock] Installed {pkg} ({version}) successfully.")
        return True

    stderr = result.stderr.decode().strip()
    print(f"[pylock] Installation error: {stderr}")


    if not _is_retry:
        guessed = guess_distribution_name(package)
        if guessed and guessed.lower() != package.lower():
            print(f"[pylock] Trying again with guessed pip name: {guessed}")
            return install_package(guessed, version, _is_retry=True)

    raise RuntimeError(f"[pylock] Failed to install {package}")


def ensure_package(module_name: str, version: str = None):
    try:
        __import__(module_name)
        return True
    except ImportError:
        print(f"[pylock] {module_name} not found. Attempting install...")
        return install_package(module_name, version)

def load_known_depmap():
    try:
        with resources.files("pydepguard.pylock").joinpath("known_deps.pydepcache").open("r", encoding="utf-8") as f:
            return json.load(f)
    except Exception as e:
        print(f"[pylock.WARN] Failed to load known dependency map: {e}")
        return {}
    

def extract_installed_version(stdout: bytes, package_name: str) -> str | None:
    text = stdout.decode("utf-8")
    pattern = rf"Successfully installed {re.escape(package_name)}-([\w\.\-]+)"
    match = re.search(pattern, text)
    if match:
        return match.group(1)
    return None


import os
import json
import tempfile
from pathlib import Path
from pydepguard.pylock.lockfile import LockfileManager

def test_lockfile_write_and_read():
    deps = {
        'requests': {
            'version': '2.31.0',
            'origin': 'script.py:10',
            'tree': ['urllib3', 'certifi']
        }
    }
    with tempfile.NamedTemporaryFile(suffix=".py", delete=False) as tmp:
        script_path = tmp.name

    lm = LockfileManager(script_path)

    try:
        if os.path.exists(lm.lockfile_path):
            os.remove(lm.lockfile_path)

        lm.save(deps)
        assert os.path.exists(lm.lockfile_path)
        assert lm.exists() is True

        data = lm.load()
        assert 'meta' in data
        assert 'deps' in data
        assert data['deps'] == deps
        assert data['meta']['script'] == Path(script_path).stem
        assert 'last_modified' in data['meta']
        assert 'saved_on' in data['meta']

    finally:
        if os.path.exists(lm.lockfile_path):
            os.remove(lm.lockfile_path)
        if os.path.exists(script_path):
            os.remove(script_path)

def test_lockfile_exists_false():
    with tempfile.NamedTemporaryFile(suffix=".py", delete=False) as tmp:
        script_path = tmp.name
    os.remove(script_path)

    lm = LockfileManager(script_path)
    if os.path.exists(lm.lockfile_path):
        os.remove(lm.lockfile_path)

    assert lm.exists() is False


from pydepguard.pylock.utils import enrich_dependencies
from pydepguard.pylock.depscan import ImportReference
from pathlib import Path

def test_enrich_known_package():
    ref = ImportReference(module='requests', file='app.py', line=5, import_type='import')
    enriched = enrich_dependencies([ref])

    assert 'requests' in enriched
    info = enriched['requests']
    assert info['version'] != 'unknown'
    assert 'urllib3' in info['tree']
    assert info['origin'] == 'app.py:5'

def test_enrich_unknown_package():
    ref = ImportReference(module='thisshouldnotexist1234', file='script.py', line=1, import_type='import')
    enriched = enrich_dependencies([ref])

    assert 'thisshouldnotexist1234' in enriched
    info = enriched['thisshouldnotexist1234']
    assert info['version'] == 'unknown'
    assert info['tree'] == []
    assert info['origin'] == 'script.py:1'



import tempfile
from pathlib import Path
from pydepguard.pylock.depscan import scan_script_for_imports, ImportReference

def write_temp_script(code: str) -> Path:
    with tempfile.NamedTemporaryFile(mode="w", suffix=".py", delete=False) as tmp:
        tmp.write(code)
        return Path(tmp.name)

def print_refs(label: str, results: list[ImportReference]):
    print(f"\n[{label}] Found {len(results)} imports:")
    for ref in results:
        print(f"  - {ref.import_type.upper()}: {ref.module} (line {ref.line}) symbols={ref.imported_symbols}")

def test_basic_import_scan():
    code = "import os\nimport sys\n"
    tmp_path = write_temp_script(code)

    results, _ = scan_script_for_imports(tmp_path)
    print_refs("BASIC IMPORT", results)

    modules = {r.module for r in results}
    assert {'os', 'sys'}.issubset(modules)
    assert all(isinstance(r, ImportReference) for r in results)

def test_from_import_multiple_symbols():
    code = "from math import sqrt, pow"
    tmp_path = write_temp_script(code)

    results, _ = scan_script_for_imports(tmp_path)
    print_refs("FROM IMPORT MULTIPLE", results)

    assert len(results) == 1
    assert results[0].module == 'math'
    assert sorted(results[0].imported_symbols) == ['pow', 'sqrt']

def test_import_with_alias():
    code = "import pandas as pd"
    tmp_path = write_temp_script(code)

    results, _ = scan_script_for_imports(tmp_path)
    print_refs("IMPORT WITH ALIAS", results)

    assert len(results) == 1
    assert results[0].module == 'pandas'
    assert results[0].imported_symbols == ['pd']

def test_relative_import():
    code = "from . import localmod"
    tmp_path = write_temp_script(code)

    results, _ = scan_script_for_imports(tmp_path)
    print_refs("RELATIVE IMPORT", results)

    assert len(results) == 1
    assert results[0].module == '.'
    assert 'localmod' in results[0].imported_symbols

def test_dynamic_import_builtin():
    code = "__import__('json')"
    tmp_path = write_temp_script(code)

    results, _ = scan_script_for_imports(tmp_path)
    print_refs("DYNAMIC BUILTIN", results)

    assert len(results) == 1
    assert results[0].module == 'json'
    assert results[0].import_type == 'dynamic'

def test_importlib_dynamic_import():
    code = "import importlib\nimportlib.import_module('collections')"
    tmp_path = write_temp_script(code)

    results, _ = scan_script_for_imports(tmp_path)
    print_refs("IMPORTLIB DYNAMIC", results)

    dynamic = [r for r in results if r.import_type == 'dynamic']
    assert any(r.module == 'collections' for r in dynamic)

def test_bad_syntax_graceful_fail():
    code = "def broken(:"
    tmp_path = write_temp_script(code)

    results, _ = scan_script_for_imports(tmp_path)
    print_refs("BAD SYNTAX", results)

    assert results == []

def test_import_without_alias():
    code = "import requests"
    tmp_path = write_temp_script(code)

    results, _ = scan_script_for_imports(tmp_path)
    assert len(results) == 1
    assert results[0].imported_symbols == []

def test_relative_import_level_2():
    code = "from .. import configtools"
    tmp_path = write_temp_script(code)

    results, _ = scan_script_for_imports(tmp_path)
    assert len(results) == 1
    assert results[0].module == '..'
    assert results[0].imported_symbols == ['configtools']


def test_from_import_no_level():
    code = "from flask import Flask"
    tmp_path = write_temp_script(code)

    results, _ = scan_script_for_imports(tmp_path)
    assert results[0].module == 'flask'

def test_import_module_wrong_object():
    code = "getlib().import_module('something')"
    tmp_path = write_temp_script(code)

    results, _ = scan_script_for_imports(tmp_path)
    print_refs("IMPORTLIB INVALID STRUCTURE", results)

    assert all(r.module != 'something' for r in results)

def test_import_from_without_module():
    code = "from  import foo"
    tmp_path = write_temp_script(code)

    results, _ = scan_script_for_imports(tmp_path)
    print_refs("FROM IMPORT MISSING MODULE", results)

    assert results == []

def test_relative_import_with_module():
    code = "from ..utils import foo"
    tmp_path = write_temp_script(code)

    results, _ = scan_script_for_imports(tmp_path)
    print_refs("RELATIVE IMPORT WITH MODULE", results)

    assert len(results) == 1
    assert results[0].module == '..utils'
    assert results[0].imported_symbols == ['foo']

def test_call_name_not_import():
    code = "do_something('foo')"
    tmp_path = write_temp_script(code)

    results, _ = scan_script_for_imports(tmp_path)
    print_refs("CALL NAME NOT IMPORT", results)

    assert all(r.import_type != 'dynamic' for r in results)

def test_import_module_wrong_id():
    code = """
class Dummy:
    id = 'not_importlib'
Dummy().import_module('notreal')
"""
    tmp_path = write_temp_script(code)
    results, _ = scan_script_for_imports(tmp_path)
    print_refs("IMPORT MODULE WRONG ID", results)
    assert all(r.module != 'notreal' for r in results)

def test_import_module_non_string_arg():
    code = "import importlib\nimportlib.import_module(42)"
    tmp_path = write_temp_script(code)

    results, _ = scan_script_for_imports(tmp_path)
    print_refs("IMPORTLIB NON-STRING ARG", results)

    assert all(r.module != '42' for r in results)

def test_dunder_import_variable_arg():
    code = '''
mod = "json"
__import__(mod)
'''
    tmp_path = write_temp_script(code)

    results, _ = scan_script_for_imports(tmp_path)
    print_refs("DUNDER IMPORT VAR ARG", results)

    assert all(r.import_type != 'dynamic' for r in results)

def test_dunder_import_empty_args():
    code = "__import__()"
    tmp_path = write_temp_script(code)

    results, _ = scan_script_for_imports(tmp_path)
    print_refs("DUNDER IMPORT EMPTY ARGS", results)
    assert results == []

def test_importlib_no_args():
    code = "import importlib\nimportlib.import_module()"
    tmp_path = write_temp_script(code)

    results, _ = scan_script_for_imports(tmp_path)
    print_refs("IMPORTLIB NO ARGS", results)

    assert any(r.module == 'importlib' and r.import_type == 'import' for r in results)
    assert all(r.import_type != 'dynamic' for r in results)


import json
import os
from pathlib import Path
from datetime import datetime, timezone

class LockfileManager:
    def __init__(self, script_path):
        self.script_path = Path(script_path)
        self.script_name = self.script_path.stem
        self.lockfile_name = f"{self.script_name}_dep.lck"
        self.lockfile_dir = self.script_path.parent / ".pylock"
        self.lockfile_dir.mkdir(exist_ok=True)
        self.lockfile_path = self.lockfile_dir / self.lockfile_name
        self.lockfile = None

    def exists(self):
        return self.lockfile_path.exists()

    def load(self):
        with open(self.lockfile_path, 'r') as f:
            self.lockfile = json.load(f)
        return self.lockfile

    def save(self, deps_info):
        enriched_deps = {}
        for dep, info in deps_info.items():
            enriched_deps[dep] = {
                'version': info.get('version', 'unknown'),
                'origin': info.get('origin', 'unknown'),
                'tree': info.get('tree', [])
            }

        lockfile_content = {
            'meta': {
                'script': self.script_name,
                'path': str(self.script_path),
                'saved_on': datetime.now(timezone.utc).isoformat() + 'Z',
                'last_modified': int(os.path.getmtime(self.script_path))
            },
            'deps': enriched_deps
        }
        with open(self.lockfile_path, 'w') as f:
            json.dump(lockfile_content, f, indent=4)
        print(f"Generated new lockfile: {self.lockfile_path}")


import sys
import sysconfig

def is_stdlib_module(module: str) -> bool:
    return module in sys.builtin_module_names or module in sysconfig.get_paths()['stdlib']

def resolve_top_level_package(module: str) -> str:
    return module.split('.')[0]



import os
import json
import tempfile
from pathlib import Path
from pydepguard.pylock.cli import main as pylock_main
from pydepguard.pylock.lockfile import LockfileManager
import subprocess
import sys


def test_cli_generate_validate_run(monkeypatch, capsys):
    code = (
        "import requests\n"
        "import json\n"
        "from math import sqrt\n"
        "print('PyLock CLI Test Success')"
    )
    with tempfile.NamedTemporaryFile(mode="w", suffix=".py", delete=False) as tmp:
        tmp.write(code)
        script_path = tmp.name

    lockfile_path = Path(script_path).parent / ".pylock" / f"{Path(script_path).stem}_dep.lck"

    try:
        sys.argv = ["pylock", script_path, "--generate"]
        pylock_main()
        assert lockfile_path.exists()

        sys.argv = ["pylock", script_path, "--validate"]
        monkeypatch.setattr("builtins.input", lambda _: "yes")
        pylock_main()

        sys.argv = ["pylock", script_path, "--run"]
        pylock_main()
        out = capsys.readouterr().out
        print(out)
        assert "PyLock CLI Test Success" in out

        data = json.loads(lockfile_path.read_text())
        print(json.dumps(data, indent=4))
        assert "requests" in data["deps"]
        assert "version" in data["deps"]["requests"]
        assert "tree" in data["deps"]["requests"]

    finally:
        if lockfile_path.exists():
            os.remove(lockfile_path)
        if Path(script_path).exists():
            os.remove(script_path)
        pylock_dir = Path(script_path).parent / ".pylock"
        if pylock_dir.exists() and not any(pylock_dir.iterdir()):
            pylock_dir.rmdir()

def test_cli_help_output(capsys):
    sys.argv = ["pylock", "dummy.py"]
    try:
        pylock_main()
    except SystemExit:
        pass

    captured = capsys.readouterr()
    assert "[pylock] Error: File not found: dummy.py" in captured.err

def test_cli_explicit_help(capsys):
    sys.argv = ["pylock", "--help"]
    try:
        pylock_main()
    except SystemExit:
        pass

    captured = capsys.readouterr()
    assert "PyLock: A gatekeeper dependency validator for Python scripts" in captured.out
    assert "Usage: pylock script.py" in captured.out
    assert "--generate" in captured.out
    assert "--validate" in captured.out
    assert "--run" in captured.out

def test_cli_no_flags_shows_help(capsys):
    sys.argv = ["pylock"]
    try:
        pylock_main()
    except SystemExit:
        pass

    captured = capsys.readouterr()
    assert "PyLock: A gatekeeper dependency validator for Python scripts" in captured.out
    assert "Usage: pylock script.py" in captured.out

def test_cli_print_help_if_no_action(capsys):
    code = "print('Hi')"
    with tempfile.NamedTemporaryFile(mode="w", suffix=".py", delete=False) as tmp:
        tmp.write(code)
        script_path = tmp.name

    try:
        sys.argv = ["pylock", script_path]
        try:
            pylock_main()
        except SystemExit:
            pass

        captured = capsys.readouterr()
        assert "[pylock] No action specified" in captured.out
        assert "Usage: pylock script.py" in captured.out

    finally:
        if Path(script_path).exists():
            os.remove(script_path)

def test_cli_validate_missing_lockfile(capsys):
    code = "print('just a test')"
    with tempfile.NamedTemporaryFile(mode="w", suffix=".py", delete=False) as tmp:
        tmp.write(code)
        script_path = tmp.name

    pylock_dir = Path(script_path).parent / ".pylock"
    lockfile_path = pylock_dir / f"{Path(script_path).stem}_dep.lck"
    if lockfile_path.exists():
        os.remove(lockfile_path)

    try:
        sys.argv = ["pylock", script_path, "--validate"]
        try:
            pylock_main()
        except SystemExit:
            pass

        captured = capsys.readouterr()
        assert "No lockfile found" in captured.err
        assert "--generate first" in captured.err

    finally:
        if Path(script_path).exists():
            os.remove(script_path)
        if pylock_dir.exists() and not any(pylock_dir.iterdir()):
            pylock_dir.rmdir()




import os
import tempfile
import subprocess
from pydepguard.pylock.runner import execute_script

def test_execute_script_runs_successfully(capsys):
    code = "print(\"Hello from test script\")"
    with tempfile.NamedTemporaryFile(mode="w", suffix=".py", delete=False) as tmp:
        tmp.write(code)
        script_path = tmp.name

    try:
        execute_script(script_path)
        captured = capsys.readouterr()
        assert "Hello from test script" in captured.out
    finally:
        if os.path.exists(script_path):
            os.remove(script_path)

def test_execute_script_failure(capsys):
    code = "raise ValueError(\"Test error\")"
    with tempfile.NamedTemporaryFile(mode="w", suffix=".py", delete=False) as tmp:
        tmp.write(code)
        script_path = tmp.name

    try:
        execute_script(script_path)
        captured = capsys.readouterr()
        assert "Test error" in captured.err
        assert "return code" in captured.out.lower()
    finally:
        if os.path.exists(script_path):
            os.remove(script_path)


import importlib.metadata
import subprocess
from .package_handler import ensure_package, install_package

def resolve_installed_package_info(package_name: str) -> dict:
    try:
        version = importlib.metadata.version(package_name)
        return {'available': True, 'version': version, 'source': 'importlib'}
    except importlib.metadata.PackageNotFoundError:
        pass

    result = subprocess.run(['pip', 'show', package_name], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    if result.returncode == 0:
        output = result.stdout.decode('utf-8')
        version = None
        for line in output.splitlines():
            if line.lower().startswith("version:"):
                version = line.split(": ", 1)[1].strip()
                break
        return {'available': True, 'version': version, 'source': 'pip_show'}

    return {'available': False, 'version': None, 'source': 'none'}

def check_package_availability(package, expected_version=None):
    info = resolve_installed_package_info(package)

    return {
        'available': info['available'],
        'version_matches': (expected_version is None or expected_version == info['version']),
        'version': info['version'],
        'source': info['source']
    }

def validate_environment(lockfile, *, strict=True, interactive=True, on_error='abort', fix_missing=False):
    if not isinstance(lockfile, dict) or 'deps' not in lockfile:
        raise ValueError("[pylock] Invalid lockfile format: 'deps' key missing")

    for dep, info in lockfile['deps'].items():
        try:
            result = check_package_availability(dep, info.get('version'))
        except Exception as e:
            print(f"[pylock] Error checking {dep}: {e}")
            if on_error == 'abort':
                raise RuntimeError(f"[pylock] Dependency check failed for {dep}")
            elif on_error == 'warn':
                continue
            else: 
                continue

        if not result['available']:
            msg = f"[pylock] Missing required package: {dep}"
            if fix_missing:
                try:
                    ensure_package(dep, info.get('version'))
                    continue
                except Exception as e:
                    print(f"[pylock.WARN] Auto-install failed: {e}")
            if on_error == 'abort': 
                raise RuntimeError(msg)
            elif on_error == 'warn':
                print(f"WARNING: {msg}")
                continue
            else:
                continue

        if strict and not result['version_matches']:
            msg = (f"[pylock] Version mismatch for {dep}: "
                   f"expected {info['version']}, found {result['version']}")
            print(msg)
            if interactive:
                try:
                    response = input("Continue anyway? (yes/no): ")
                except KeyboardInterrupt:
                    print("\n[pylock] Aborted by user.")
                    raise SystemExit(130)
                if response.strip().lower() != 'yes':
                    raise RuntimeError("[pylock] Validation aborted due to version mismatch.")
            else:
                if on_error == 'abort':
                    raise RuntimeError("[pylock] Validation failed due to version mismatch.")
                elif on_error == 'warn': 
                    print(f"WARNING: {msg}")
                    continue
                else: 
                    continue

    print("[pylock] Environment validation passed.")


import argparse
import sys
import json
from pathlib import Path
from .depscan import scan_script_for_imports
from .lockfile import LockfileManager
from .validator import validate_environment
from .runner import execute_script
from .utils import enrich_dependencies

from time import time

gtime = time()


def main():
    parser = argparse.ArgumentParser(
        description="PyLock: A gatekeeper dependency validator for Python scripts.\n"
                    "This tool scans Python scripts for imports, generates lockfiles, and validates dependencies.\n"
                    "Version 3.0.4 - Made by 0xIkari\n"
                    "Part of the PyDepGuard project\n"
                    "Usage: pylock script.py [options]\n\n"
                    "Options:\n"
                    "  --generate         Generate or overwrite per-file lockfile\n"
                    "  --validate         Validate environment against lockfile\n"
                    "  --run              Run the script if validation passes\n"
                    "  --strict           Enable strict version matching\n"
                    "  --non-interactive  Disable user input (e.g., for CI/CD)\n"
                    "  --on-error         Set behavior on errors: 'abort', 'warn', or 'skip'\n"
                    "  --fix-missing      Install any missing dependencies as found during AST or locklife read\n",
        formatter_class=argparse.RawTextHelpFormatter
    )
    parser.add_argument('script', nargs='?', help="Script to check and run")
    parser.add_argument('-g', '--generate', action='store_true')
    parser.add_argument('-v', '--validate', action='store_true')
    parser.add_argument('-r', '--run', action='store_true')
    parser.add_argument('--strict', action='store_true')
    parser.add_argument('--non-interactive', action='store_true')
    parser.add_argument('--on-error', choices=['abort', 'warn', 'skip'], default='abort')
    parser.add_argument('--fix-missing', action='store_true')

    args = parser.parse_args()

    if not args.script:
        parser.print_help()
        sys.exit(1)

    script_path = Path(args.script)
    if not script_path.exists():
        print(f"[pylock] Error: File not found: {script_path}", file=sys.stderr)
        sys.exit(1)

    lm = LockfileManager(script_path)

    if args.generate:
        print("[pylock] Scanning for imports...")
        imports, unbound_symbols = scan_script_for_imports(script_path)
        print(f"[pylock] Found {len(unbound_symbols)} unbound symbols.")
        unbound_findings = list()
        for sym in unbound_symbols:
            unbound_findings.append(f"File: {sym.file} Module: {sym.name} Line Number: {sym.line})")
            print(f"[pylock.CRIT] Unbound Symbol: {sym.name} at {sym.file}:{sym.line} - Add `import {sym.name}` to {sym.file} resolve.")
        deps = enrich_dependencies(imports)
        lm.save(deps)
        print(f"[pylock] Lockfile generated for {script_path.name} with {len(deps)} dependencies.")
        print(f"[pylock.DBG] Total Time Spent: {time() - gtime:.8f} seconds")
        print(f"If this helped you save time, please star or sponsor me: https://github.com/nuclear-treestump/pylock-dependency-lockfile")
        return

    if args.validate or args.run:
        if not lm.exists():
            print(f"[pylock] Error: No lockfile found for {script_path.name}. Please run with --generate first.", file=sys.stderr)
            sys.exit(1)

        lockfile = lm.load()
        validate_environment(
            lockfile,
            strict=args.strict,
            interactive=not args.non_interactive,
            on_error=args.on_error,
            fix_missing=args.fix_missing
        )
        if args.run:
            execute_script(args.script)
        print(f"[pylock.DBG] Total Time Spent: {time() - gtime:.8f} seconds")
        print(f"If this helped you save time, please star or sponsor me: https://github.com/nuclear-treestump/pylock-dependency-lockfile")
        return

    print("[pylock] No action specified. Use --generate, --validate, or --run.\n")
    parser.print_help()



import logging

def config(policy="strict", fill_missing=True, venv=1, nx=True, os_check=True):
    logging.basicConfig(level=logging.INFO)
    logging.info(f"PyLock Configured: policy={policy}, fill_missing={fill_missing}, venv={venv}, nx={nx}, os_check={os_check}")

    


import subprocess
import sys

def execute_script(script_path):
    print(f"Running {script_path}...")
    process = subprocess.Popen([sys.executable, script_path], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout, stderr = process.communicate()

    if stdout:
        print(stdout.decode(), end='')
    if stderr:
        print(stderr.decode(), end='', file=sys.stderr)

    rc = process.returncode
    if rc != 0:
        print(f"Script exited with return code {rc}")


KNOWN_DEP_MAP = {
  "yaml": "pyyaml",
  "bs4": "beautifulsoup4",
  "cv2": "opencv-python",
  "PIL": "pillow",
  "sklearn": "scikit-learn",
  "Crypto": "pycryptodome",
  "Image": "pillow",
  "lxml": "lxml",
  "win32com": "pywin32"
}

KNOWN_TRANSITIVE = {
    "pandas.read_excel": ["openpyxl", "xlrd"],
    "pandas.read_html": ["lxml", "html5lib", "bs4"],
    "sqlalchemy.create_engine": ["sqlite3", "psycopg2", "mysqlclient"],
    "yaml.safe_load": ["pyyaml"],
    "matplotlib.use": ["matplotlib"],
    "pydantic.BaseModel": ["pydantic"],
}


import importlib.metadata
from importlib.util import find_spec
from pathlib import Path
import sysconfig
import sys
import ast
import re
from typing import List, Dict
from .depscan import ImportReference


def strip_extras(requirement: str) -> str:
    return re.split(r"[<>=!~\[]", requirement)[0].strip()



def enrich_dependencies(imports: List[ImportReference]) -> Dict[str, dict]:
    enriched = {}

    for ref in imports:
        top_package = ref.module.split('.')[0]
        if is_stdlib_module(top_package):
            continue

        try:
            dist = importlib.metadata.distribution(top_package)
            version = dist.version
            requires = dist.requires or []
            transitive = [strip_extras(r) for r in requires if r and not r.startswith('extra')]
        except importlib.metadata.PackageNotFoundError:
            version = "unknown"
            transitive = []

        enriched[top_package] = {
            'version': version,
            'origin': f"{ref.file}:{ref.line}",
            'tree': sorted(set(transitive))
        }

    return enriched

def is_stdlib_module(module: str) -> bool:
    import importlib.util
    spec = importlib.util.find_spec(module)
    return spec is not None and 'site-packages' not in (spec.origin or '')

def resolve_symbol_dependencies(symbol_fqname: str) -> List[str]:
    parts = symbol_fqname.split('.')
    for i in reversed(range(1, len(parts))):
        module_name = ".".join(parts[:i])
        attr = parts[i]

        try:
            spec = find_spec(module_name)
            if not spec or not spec.origin or not spec.origin.endswith(".py"):
                continue

            module_path = Path(spec.origin)
            with module_path.open("r", encoding="utf-8") as f:
                module_ast = ast.parse(f.read(), filename=str(module_path))
        except Exception as e:
            continue

        for node in ast.walk(module_ast):
            if isinstance(node, ast.FunctionDef) and node.name == attr:
                deps = set()
                for inner in ast.walk(node):
                    if isinstance(inner, ast.Import):
                        for alias in inner.names:
                            deps.add(alias.name.split('.')[0])
                    elif isinstance(inner, ast.ImportFrom):
                        if inner.module:
                            deps.add(inner.module.split('.')[0])
                return sorted(deps)
    return []
